/***
 * Excerpted from "Modern Systems Programming with Scala Native",
 * published by The Pragmatic Bookshelf.
 * Copyrights apply to this code. It may not be used to create training material,
 * courses, books, articles, and the like. Contact us if you are in doubt.
 * We make no guarantees that this code is fit for any purpose.
 * Visit http://www.pragmaticprogrammer.com/titles/rwscala for more book information.
***/
import scala.scalanative.unsafe._
import scala.scalanative.libc._
import stdlib._
import string._
import stdio._

final case class WrappedArray[T](var data:Ptr[T], var used:Int, var capacity:Int)

object main {
  type NGramData = CStruct4[CString, Int, Int, Int]

  def makeWrappedArray(size:Int):WrappedArray[NGramData] = {
    val data = malloc(size * sizeof[NGramData]).asInstanceOf[Ptr[NGramData]]
    return WrappedArray[NGramData](data, 0, size)
  }

  def growWrappedArray(array:WrappedArray[NGramData], size:Int):Unit = {
    val new_capacity = array.capacity + size
    val new_size = new_capacity * sizeof[NGramData]
    val new_data = realloc(array.data.asInstanceOf[Ptr[Byte]], new_size)
    array.data = new_data.asInstanceOf[Ptr[NGramData]]
    array.capacity = new_capacity
  }

  val by_count = new CFuncPtr2[Ptr[Byte],Ptr[Byte],Int] {
    def apply(p1:Ptr[Byte], p2:Ptr[Byte]):Int = {
      val ngram_ptr_1 = p1.asInstanceOf[Ptr[NGramData]]
      val ngram_ptr_2 = p2.asInstanceOf[Ptr[NGramData]]
      val count_1 = ngram_ptr_1._2
      val count_2 = ngram_ptr_2._2
      return count_2 - count_1
    }
  }

  val line_buffer = malloc(1024)
  val temp_word_buffer = malloc(1024)
  val block_size = 65536 * 16 // ~ 1MB - too big?

def main(args:Array[String]):Unit = {
  val array = makeWrappedArray(block_size)

  val read_start = System.currentTimeMillis()
  val lines_read = readAllLines(stdin,array)
  val read_elapsed = System.currentTimeMillis() - read_start
  println(s"""done. read $lines_read lines, ${array.used} unique words.
          $read_elapsed ms""")

  val sort_start = System.currentTimeMillis()
  qsort.qsort(array.data.asInstanceOf[Ptr[Byte]], array.used,
              sizeof[NGramData], by_count)
  val sort_elapsed = System.currentTimeMillis() - sort_start
  stdio.printf(c"sorting done in %d ms\n", sort_elapsed)

  val to_show = if (array.used <= 20) array.used else 20
  for (i <- 0 until to_show) {
    stdio.printf(c"word n: %s %d\n", (array.data + i)._1, (array.data + i)._2)
  }

  println(c"done")
}

def readAllLines(fd:Ptr[stdio.FILE], array:WrappedArray[NGramData]):Long = {
  var lines_read = 0L
  while (stdio.fgets(line_buffer, 1024, fd) != null) {
    if (array.used >= (array.capacity - 1)) {
      growWrappedArray(array, block_size)
    }
    parseAndCompare(line_buffer, array)
    lines_read += 1
    if (lines_read % 10000000 == 0) {
      stdio.printf(c"read %d lines, %d unique words so far\n", lines_read,
                   array.used)
    }
  }
  return lines_read
}

def parseAndCompare(line:CString, array:WrappedArray[NGramData]):Unit = {
  val temp_item = stackalloc[NGramData]
  temp_item._1 = temp_word_buffer
  val next_item = array.data + array.used
  scan_item(line, temp_item)
  if (array.used == 0) {
    add_new_item(temp_item, next_item)
    array.used += 1
  } else {
    val prev_item = array.data + (array.used - 1)
    if (is_item_new(temp_item, prev_item) != 0) {
      add_new_item(temp_item, next_item)
      array.used += 1
    } else {
      accumulate_item(temp_item, prev_item)
    }
  }
}

def scan_item(line:CString, temp_item:Ptr[NGramData]):Boolean = {
  val temp_word = temp_item._1
  val temp_count = temp_item.at2
  val temp_year = temp_item.at3
  val temp_doc_count = temp_item.at4

  val sscanf_result = stdio.sscanf(line, c"%1023s %d %d %d\n",
    temp_word, temp_year, temp_count, temp_doc_count)
  if (sscanf_result < 4) {
    throw new Exception("input error")
  }

  return true
}

def is_item_new(temp_item:Ptr[NGramData], prev_item:Ptr[NGramData]):Int = {
  strcmp(temp_item._1, prev_item._1)
}

def add_new_item(temp_item:Ptr[NGramData], 
                 next_item:Ptr[NGramData]):Unit = {
  val temp_word = temp_item._1
  val new_word_length = strlen(temp_word)
  val new_word_buffer = malloc(new_word_length + 1)

  strncpy(new_word_buffer, temp_word, new_word_length)
  new_word_buffer(new_word_length) = 0

  next_item._1 = new_word_buffer
  next_item._2 = temp_item._2
  next_item._3 = temp_item._3
  next_item._4 = temp_item._4
}

def accumulate_item(temp_item:Ptr[NGramData], 
                    prev_item:Ptr[NGramData]):Unit = {
  prev_item._2 = prev_item._2 + temp_item._2
}
}

@extern
object qsort {
  def qsort(data:Ptr[Byte],
          num:Int,
          size:Long,
          comparator:CFuncPtr2[Ptr[Byte], Ptr[Byte], Int]):Unit = extern

}
